(*********************************************************************)
(*                         entity.n                                  *)	
(*-------------------------------------------------------------------*)
(*  By Hyun-Goo Kang                                                 *)
(*-------------------------------------------------------------------*)
(*  Implement entity related processing                              *)
(*   - parameter, general entity environment definition              *) 
(*   - entity environment lookup, insert                             *)   
(*   - check kind, get string, get uri of a external entity          *)
(*   - nomalization of entity string                                 *)   
(*   - notification routine                                          *)   
(*-------------------------------------------------------------------*)
(*  To do                                                            *)
(*-------------------------------------------------------------------*)
(*  1. Charater reference replacement should support 2 byte encoding *) 
(*   - it's not directly possible in nML because nML don't support   *)
(*     two byte charater in a string                                 *)
(*  2. Public literal processing : optional                          *)
(*    e.g) public literal for HTML                                   *)
(*********************************************************************)

signature ENTITY_PROCESSING =
sig 
  type notation = EMPTY_NOTATION | N of string
  type reference = SYSTEM_REF of string | PUBLIC_REF of string * string
  type p_entity = P_STR of string | P_REF of reference
  type g_entity = G_STR of string | G_REF of reference * notation
  
  val pe_entity_env_insert : string -> p_entity -> unit
  val ge_entity_env_insert : string -> g_entity -> unit
  val pe_entity_env_lookup : string -> p_entity
  val ge_entity_env_lookup : string -> g_entity
  val pe_is_empty : p_entity -> bool
  val ge_is_empty : g_entity -> bool
  val ge_is_external : g_entity -> bool
  val ge_is_unparsed : g_entity -> bool
  val ge_extract_notation : g_entity -> string
  val ge_extract_string : g_entity -> string
  val pe_is_external : p_entity -> bool
  val pe_extract_string : p_entity -> string
  val pe_get_uri : p_entity -> string
  val ge_get_uri : g_entity -> string
  val normalize_attr_value : string -> string 
  val normalize_entity_value : string -> string 
  val notify : g_entity -> unit 
end  


structure Entity_proccesing : ENTITY_PROCESSING = 
struct

open Global
open Nxml_errors
open System_Processing

exception Exn_entity of string

// an adhoc approach
// fix it to production text decl
fun eliminate_text_decl(src_str) = let  
  val len = String.length src_str
  val pat_str = "<\?xml[^\255]*\?>"
  val pat = Str.regexp pat_str
  val s_pos = Str.search_forward pat src_str 0 
  val e_pos = Str.match_end()  
  val prefix = String.sub src_str 0 s_pos 
  val suffix = String.sub src_str e_pos (len-e_pos)
in
  prefix ^ suffix
end handle Not_found => src_str
         | _ => raise Exn_entity("error in eliminating text decl ")
    
type notation = EMPTY_NOTATION | N of string
type reference = SYSTEM_REF of string | PUBLIC_REF of string * string
type p_entity = P_STR of string | P_REF of reference
type g_entity = G_STR of string | G_REF of reference * notation

(* pe environment *)
type id = string
type env = (id * g_entity) list
type penv = (id * p_entity) list
val emptyGE : env = [] 
val emptyPE : penv = []

val pe_env = ref emptyPE
fun pe_entity_env_insert id (P_STR s) = 
      pe_env := (id,P_STR(s))::(!pe_env)
  | pe_entity_env_insert id (P_REF(r)) = 
    case r of 
      SYSTEM_REF(s) => 
      let val absolute_uri = absolute_path (current_URI()) s
          val _ = if (!Global.debug_tag) then 
                  (print_string ("\nInserting a p_entity : system"^s);
                  print_newline();
                  print_string ("absolute_uri="^absolute_uri))
      in  pe_env := (id,P_REF(SYSTEM_REF(absolute_uri)))::(!pe_env) end
    | PUBLIC_REF(s1,s2) => 
      let val absolute_uri = absolute_path (current_URI()) s2
      in  pe_env := (id,P_REF(PUBLIC_REF(s1,absolute_uri)))::(!pe_env) end
fun pe_entity_env_lookup id = 
let val (_,p_entity) = List.find (fn (x,_) => x=id) !pe_env 
 		      handle Not_found => 
				raise XMLParserError(
				        Other_errors ("undefined p-entity reference:"^id,!line_no))
    		           | x => raise x	
in p_entity end

val ge_env = ref emptyGE

fun ge_entity_env_insert id (G_STR s) =  
       ge_env := (id,G_STR(s))::(!ge_env)
  | ge_entity_env_insert id (G_REF(r,n)) = 
    case r of 
      SYSTEM_REF(s) => 
      let val absolute_uri = absolute_path (current_URI()) s
      in  ge_env := (id,G_REF(SYSTEM_REF(absolute_uri),n))::(!ge_env) end
    | PUBLIC_REF(s1,s2) => 
      let val absolute_uri = absolute_path (current_URI()) s2
      in  ge_env := (id,G_REF(PUBLIC_REF(s1,absolute_uri),n))::(!ge_env) end
fun ge_entity_env_lookup id = 
let val (_,g_entity) = List.find (fn (x,_) => x=id) !ge_env 
    		      handle Not_found => 
				raise XMLParserError(
				        Other_errors ("undefiend g-entity reference:"^id,!line_no))
    		           | x => raise x	
in g_entity end

fun ge_is_external(G_REF(_,_)) = true
  | ge_is_external(_) = false
fun ge_is_unparsed(G_REF(_,N(_))) = true
  | ge_is_unparsed(_) = false
fun ge_extract_string (G_STR(s)) = s
  | ge_extract_string (G_REF(SYSTEM_REF(s),EMPTY_NOTATION)) = 
        string_from_system_literal(s)
  | ge_extract_string (G_REF(PUBLIC_REF(s1,s2),EMPTY_NOTATION)) = 
        string_from_system_literal(s2)
  | ge_extract_string (G_REF(SYSTEM_REF(s1),N(s2))) = 
      raise Exn_entity("You can't reference unparsed ge directly")
  | ge_extract_string (G_REF(PUBLIC_REF(s1,s2),N(s3))) =
      raise Exn_entity("You can't reference unparsed ge directly")
      
fun ge_extract_notation(G_REF(SYSTEM_REF(s1),N(s2))) = s2 
  | ge_extract_notation(G_REF(PUBLIC_REF(s1,s2),N(s3))) = s3
  | ge_extract_notation(_) = 
      raise Exn_entity("You cannot extract notation from non-external general unparsed entity")

fun notify(G_REF(SYSTEM_REF(s1),N(s2))) = 
    (print_string ("Notifying system reference " ^ s1 ^ " to ndata(" ^ s2 ^ ")");
     print_newline())
  | notify(G_REF(PUBLIC_REF(s1,s2),N(s3))) = 
    (print_string ("Notifying public reference (" ^ s1 ^ "," ^ s2 ^ ") to ndata(" ^ s2 ^ ")");
     print_newline())         
  | notify(_) = raise Exn_entity("Notify on non-external gerneral unparsed entity")
        
fun pe_is_external(P_REF(_)) = true
  | pe_is_external(_) = false
fun pe_extract_string(P_STR(s)) = s  //  ƹ ref
  | pe_extract_string(P_REF(SYSTEM_REF(s))) = let //  declsep ref
      val str = string_from_system_literal(s)  // it's normalized
      val str'' = eliminate_text_decl(str)      
    in
      str''  
    end    
  | pe_extract_string(P_REF(PUBLIC_REF(s1,s2))) = let  // need expand for s1  
      val str = string_from_system_literal(s2)
      val str'' = eliminate_text_decl(str)
    in
      str''  
    end    
fun pe_is_empty(p_entity) = if (pe_extract_string p_entity = "") then true else false
fun ge_is_empty(g_entity) = if (ge_extract_string g_entity = "") then true else false

fun pe_get_uri(P_STR(s)) = raise Exn_entity("uri reference about an external pe")
  | pe_get_uri(P_REF(SYSTEM_REF(s))) = s
  | pe_get_uri(P_REF(PUBLIC_REF(s1,s2))) = s2
fun ge_get_uri(G_STR(s)) = raise Exn_entity("uri reference about an external ge")
  | ge_get_uri(G_REF(SYSTEM_REF(s),n)) = s
  | ge_get_uri(G_REF(PUBLIC_REF(s1,s2),n)) = s2  

val _ = ge_entity_env_insert "amp"  (G_STR "&#38;")
val _ = ge_entity_env_insert "lt"   (G_STR "&#60;")
val _ = ge_entity_env_insert "gt"   (G_STR ">")
val _ = ge_entity_env_insert "apos" (G_STR "'")
val _ = ge_entity_env_insert "quot" (G_STR "\"")

// * entity value parameter entity "included in literal" å  ԵǾ Ѵ.
//    - ⼭  Ϳ char ref pe ref ־ ̹ Ȯ ʴ´
//    -  entity env 鶧  expandϸ   external entity ϰ
//      ״ flat 
fun normalize_entity_value src_str  = let
  val len = String.length src_str
  val pat_str = "%[a-z A-Z \192-\214 \216-\246 \248-\255 \183 0-9 \046 \058 \045 \095]*;"
  		^"\|"^
                 "&#[0-9]+;\|&#x[a-f A-F 0-9]+;"
  val pat = Str.regexp pat_str
  val s_pos = Str.search_forward pat src_str 0 
  val e_pos = Str.match_end()
  val match = Str.matched_string src_str
  val prefix = String.sub src_str 0 s_pos 
  val suffix = String.sub src_str e_pos (len-e_pos)
in
  prefix ^
  (if ((String.sub match 0 2) = "&#") then
    let       
      val num_str = "0" ^ (String.sub match 2 ((String.length match)-3))
      val num = int_of_string num_str
    in 	
      if (num > 255) then // fix
        (print_string ("Warning: this processor don't support 2 byte character(num_str="^num_str^")");
         print_string ("         substituting it to ?\n");
         "?")
      else  
        (String.make 1 (Char.chr(num))) 
    end  
   else
     let
       val p_entity = pe_entity_env_lookup(String.sub match 1 ((String.length match)-2))
       val str = pe_extract_string p_entity
     in
       normalize_entity_value(str)
     end
  ) 
  ^ (normalize_entity_value suffix)
end handle Not_found => src_str
                 
fun normalize_attr_value src_str  = let
  val len = String.length src_str
  val pat_str_attr = "[\013 \032 \010 \009]"
                ^"\|"^ 
                "&[a-z A-Z \192-\214 \216-\246 \248-\255 \183 0-9 \046 \058 \045 \095]*;"
  		^"\|"^
                 "&#[0-9]+;\|&#x[a-f A-F 0-9]+;"                
  val pat = Str.regexp pat_str_attr
  val s_pos = Str.search_forward pat src_str 0 
  val e_pos = Str.match_end()
  val match = Str.matched_string src_str
  val prefix = String.sub src_str 0 s_pos 
  val suffix = String.sub src_str e_pos (len-e_pos)
in
  case match of
   ("\032"|"\013"|"\010"|"\009") => 
     (prefix ^ " " ^ (normalize_attr_value suffix))
  | _ => 
  (prefix ^
          (if ((String.sub match 0 2) = "&#") then
             let val num_str = "0" ^ (String.sub match 2 ((String.length match)-3))
               val num = int_of_string num_str
             in 	
                if (num > 255) then // fix
                   (print_string ("Warning: this processor don't support 2 byte character(num_str="^num_str^")\n");
                    print_string ("         substituting it to ?\n");
                   "?")
                else  
                   (String.make 1 (Char.chr(num))) 
             end
           else
             let 
               val g_id = (String.sub match 1 ((String.length match)-2))
               val g_entity = ge_entity_env_lookup(g_id)
               val _ = if ge_is_external(g_entity) then 
                       raise Exn_entity("No External Entity References : "
                                        ^ "attribute value can't contain direct or"
                                        ^ " indirect entity references to external entity")
               val s = ge_extract_string(g_entity)                         
             in
               normalize_attr_value(s)
             end
           ) 
           ^ (normalize_attr_value suffix)
          ) 
end handle Not_found => src_str
        		 
end // structure
