{
(*********************************************************************)
(*                lexer.ny                                           *)	
(*-------------------------------------------------------------------*)
(*  Initiatial development by Eun-sun Cho                            *)
(*  Finalized by Hyun-Goo Kang, ROPAS                                *)
(*-------------------------------------------------------------------*)
(*  http://ropas.kaist.ac.kr/~hgkang                                 *)
(*********************************************************************)

(*****************************************************************)
(*   Modules & some auxiliary functions used in the actions      *)
(*****************************************************************)

open Nxml_errors
open Entity_proccesing
open Parser 
open Global

(* list of rules in this lexer *)
(* dividing the lexing rule like this help lex(token)_error handling *)
(* i.e. same token has different meaning in differenct context *)
(* is it really helpful than using grammar ? hmm... think more *)
type lexers =
  	Document        // top level for external documents and external parsed entities 
      | Document_type   // type declaration : before internal declaration 
      | Content         // XML contents of elements 
      | Within_tag      // after "<name" until matching ">" 
      | Declaration     // internal declarations *)
      | XML_decl        // XML version, encoding, standalone declaration in top level 
      | Parsed_ent      // external parsed entities 
      | Pe              // external declaration of entities 
      | PI_decl         // Processing instruction declaration in top level
      | Ext_subset
   

(* check whether lexer is in top level or inside content 		*)
(* increased by 1 at each start-tag, decreased by 1 at each end-tag 	*)
val tag_count = ref 0

fun get_last_mark_str s ch =
    let val len = String.length s 
	val j = String.rindex_from s (len - 2) ch
    in  String.sub s (j+1) (len - 2 - j)   end
fun get_last_quote_str s  = get_last_mark_str s '"'
fun get_last_squote_str s  = get_last_mark_str s '\''

val text_decl_context = ref Document
val pi_decl_context = ref Document
val len_of_1st_entering_ext_subset = ref 0

} 

(***********************************************************************)
(*          Naming Frequently Used Regular Expression Section          *)
(***********************************************************************)
(* space {03} *)
(* #x9| #x20 | #xD | #xA *)
val ws = [ ' ' '\t' '\r' '\n' ]

(* Note: ISO-8859-1 charset does not have 'combining characters' 	*)

(* letter{84} *)
(* BaseChar{85} + Ideographic{86} *)
(* '\192'-'\255' : accented characters, latin characters and more 	*)
(* '\215'(muliply sign) and '\247'(divide sign) are not included 	*)
val letter = ['A'-'Z' 'a'-'z' '\192'-'\214' '\216'-'\246' '\248'-'\255']

(* Extender{89} *)
val extender = '\183'

(* Digit{88} *)
val digit = ['0'-'9']
val hexdigit = ['0'-'9' 'A'-'F' 'a'-'f']

(* NameChar{04} *)
val namechar = letter | digit | '.' | ':' | '-' | '_' | extender

(* Name{05} *)
val name = (letter | '_' | ':') namechar*
val names = (name ws+)


(* Name Token{06} *)
val nmtoken = namechar+

(* PI {16} : Not including "?>"	*)
(*  control characters '\000'-'\008' '\011'-'\012' '\014'-'\031' are 
	not included in pi_string *) 
val pi_string = [^ '?' '\000'-'\008' '\011'-'\012' '\014'-'\031']* 
		( '?' [^ '>' '\000'-'\008' '\011'-'\012' '\014'-'\031'] 
                      [^ '?' '\000'-'\008' '\011'-'\012' '\014'-'\031']* )* 
                '?'?

(* Comment {15}*)
(* except '--' *)
val comment_string = [^ '-' '\000'-'\008' '\011'-'\012' '\014'-'\031' ]*
                     ('-' [^ '-' '\000'-'\008' '\011'-'\012' '\014'-'\031' ]+ )*'-'?


(* CData {20} *)
(* excluding ']]>' *)
val character_no_rbracket = 
  [^ '&' '<' ']' '\000'-'\008' '\011'-'\012' '\014'-'\031']

val no_rbracket = [^ ']' '\000'-'\008' '\011'-'\012' '\014'-'\031']
val no_rbracket_no_rangle = [^ ']' '>'
                               '\000'-'\008' '\011'-'\012' '\014'-'\031']

val cdata_string = no_rbracket* ( "]" no_rbracket+ |
                                  "]]" ']'* no_rbracket_no_rangle no_rbracket*)* ']'*

(* Equals{25} *)
val Eq =  ws* "=" ws*

(* XML{23} *)
val XML = "XML"|"xml"

(* Version number{26} *)
val VersionNum = ['a'-'z''A'-'Z''0'-'9''_'':''-''.']+ 

(* Encoding Name{81} *)
val EncNum = ['a'-'z''A'-'Z']['A'-'Z''a'-'z''0'-'9''-''_''.']*


(****************************************************************)
(*                       lexing rules                           *)
(****************************************************************)

(* 1. scan_document: Lexer for the outermost structures 	 *)
(* - used as entrance for lexing ext_documets and parsing_entity *)
rule scan_document = parse
	"<?" XML 	
	{ text_decl_context := Document; // context to return after reading text_decl 
	  (XMLDecl_begin, XML_decl)
	}
	| "<?" name  
	{ 	let	val s = Lexing.lexeme lexbuf
			val v = String.sub s 2 ((String.length s) - 2)
		in 
		  pi_decl_context := Document;
		  (PI_begin v, PI_decl)
		end
	}	  
	| "<?" 
	{  pi_decl_context := Document;
	   (PI_NULL_begin, PI_decl) 
	}
	| 	"<!DOCTYPE"
	{ (Doctype, Document_type) }
	| 	"<!--" comment_string "-->"
	{ let val s = Lexing.lexeme lexbuf
		in 	set_line_no (s) ;(Comment s, Document) end }
	| "<!--"
	{  raise XMLParserError(Illegal("comment","top level", !line_no,
		"possible reasons are illegal control characters or double hyphens ('--') inside the comment, or missing '-->'")) }
	| "<" name
	{ 	let	val s = Lexing.lexeme lexbuf 
		in 	(tag_count:= !tag_count + 1;
			(Tag_beg (String.sub s 1 (String.length s - 1)), Within_tag ))
		end }
	| ws+
	{ 	set_line_no(Lexing.lexeme lexbuf); (Ignore,Document)(*scan_document lexbuf*)}
	| eof
	{ 	(Eof, Document) }
(* the following rules are the extended part for lexing parsed_entity 		*)
(* the above part of scan_document is shared by ext_document and parsed_entity 	*)
	| '\013' '\010'
	{ 	inc_line_no 1; (CharData "\n", Document) }
	| '\013'
	{  inc_line_no 1; (CharData "\n", Document) }
	| '\010'
	{ 	inc_line_no 1; (CharData "\n", Document) }
	| character_no_rbracket+
	{ 	let val s = Lexing.lexeme lexbuf 
		in 	set_line_no(s); (CharData s, Document) end }
	| "&#" digit+ ";"
	{ 	let val s = Lexing.lexeme lexbuf 
		in 	(CRef (int_of_string (String.sub s 2 (String.length s - 3))), Document) end }
	| "&#x" hexdigit+ ";"
	{ 	let val s = Lexing.lexeme lexbuf 
		in 	(CRef (int_of_string ("0x" ^ String.sub s 3 (String.length s - 4))), Document) end }
	| "&" name ";"
	{ let 
	    val s = Lexing.lexeme lexbuf 
	    val id = (String.sub s 1 (String.length s - 2))
	    val g_entity = ge_entity_env_lookup id
	  in
	    if (ge_is_empty(g_entity)) then
	      (Ignore, Document)
	    else
	      (ERef(id), Document)
	  end
	}	 	
	| "<![CDATA[" cdata_string "]]>"
	{ 	let val s = Lexing.lexeme lexbuf 
		in 	set_line_no(s); 
			(Cdata (String.sub s 9 (String.length s - 12)), Document) 
		end }
	| "<!"
	{  raise XMLParserError(Bad_declaration(
		"Declaration either badly formed or not allowed", "top level", !line_no) ) }
	| '<'
	{  raise XMLParserError(Illegal("<", "top level", !line_no,"")) }
	| _
	{  raise XMLParserError(Illegal(Lexing.lexeme lexbuf, "top level", !line_no,
		"content not allowed here")) }

(* 2. scan_content *)
and scan_content = parse
	   "<?" XML 
	{ text_decl_context := Content;
	  (XMLDecl_begin, XML_decl)
	}  
	| "<?" name 
	{ 	let	val s = Lexing.lexeme lexbuf
			val v = String.sub s 2 ((String.length s) - 2)
		in
		        pi_decl_context := Content; 
		        (PI_begin v, PI_decl) 
		end }
	| "<?" 
	{  (pi_decl_context := Content; (PI_NULL_begin, PI_decl)) }
	| "<!--" comment_string "-->"
	{ 	let	val s = Lexing.lexeme lexbuf
		in set_line_no s; (Comment s, Content) end }
	| "<!--"
	{ 	raise XMLParserError(Illegal("comment","content", !line_no,
		"possible reasons are illegal control characters or double hyphens ('--') inside the comment, or missing '-->'")) }
	| "<" name
	{ 	let	val s = Lexing.lexeme lexbuf 
		in (tag_count:= !tag_count + 1;
			Tag_beg (String.sub s 1 (String.length s - 1)), Within_tag )
		end }
	| "</" name
	{ 	let	val s = Lexing.lexeme lexbuf 
		in (tag_count:=!tag_count - 1;
			Tag_end (String.sub s 2 (String.length s - 2)), Within_tag )
		end }
	| "<![CDATA[" cdata_string "]]>"
	{ 	let val s = Lexing.lexeme lexbuf 
		in set_line_no s; (Cdata (String.sub s 9 (String.length s - 12)), Content)
		end }
	| "<!"
	{  raise XMLParserError(Bad_declaration(
	"Declaration either badly formed or not allowed", "content", !line_no) ) }
	| "<"
	{  raise XMLParserError(Illegal("<", "content", !line_no, 
		"The left angle bracket '<' must be written as '&lt;")) }
	| "&#" digit+ ";"
	{ 	
	  let val s = Lexing.lexeme lexbuf 
	  in (CRef (int_of_string (String.sub s 2 (String.length s - 3))), Content) end 
	}
	| "&#x" hexdigit+ ";"
	{ 	
	  let val s = Lexing.lexeme lexbuf 
	  in 
	    (CRef (int_of_string ("0x" ^ String.sub s 3 (String.length s - 4))), Content)
	  end 
	}
	| "&" name ";"
        { let 
	    val s = Lexing.lexeme lexbuf 
	    val id = (String.sub s 1 (String.length s - 2))
	    val g_entity = ge_entity_env_lookup id
	  in
	    if (ge_is_empty(g_entity)) then
	      (Ignore, Content)
	    else
	      (ERef(id), Content)
	  end
	}	 		
	
	| "&" 
	{ 	raise XMLParserError(Illegal("&", "content", !line_no, 
			"The ampersand '&' must be written as '&amp;")) }
	| '\013' '\010'
	{  inc_line_no 1; (CharData "\n", Content) }
	| '\013'
	{  inc_line_no 1;  (CharData "\n", Content) }
	| '\010'
	{  inc_line_no 1;  (CharData "\n", Content) }
	| eof
	{  (Eof, Content) }
	| "]]>" 
	{  raise XMLParserError(Illegal("]]>", "content", !line_no, 
		"The sequence ']]>' must be written as ']]&gt;")) }
	| "]"
	{  (CharData "]", Content) }
	| character_no_rbracket+
	{ 	let val s = Lexing.lexeme lexbuf 
		in (CharData s, Content) end }
	| _
	{  raise XMLParserError(Illegal(Lexing.lexeme lexbuf, "content", !line_no,"")) }

(* 3. scan_within_tag_beg: after "<name" until matching ">" *)
and scan_within_tag = parse
	ws+ 
	{  set_line_no(Lexing.lexeme lexbuf);  (Ignore, Within_tag) (*scan_within_tag lexbuf*) }
	| name
	{ 	(Name (Lexing.lexeme lexbuf ), Within_tag) }
	| '='
	{ 	(Eq, Within_tag) }
	| '"' [^ '"' ]* '"'
	{ 	let val s = Lexing.lexeme lexbuf 
			val v = String.sub s 1 (String.length s - 2) 
		in set_line_no(s); (Unparsed_string(v), Within_tag) end }
	| '"'
	{ 	raise XMLParserError(Mismatch("the second quotation mark \"", "tag", !line_no)) }
	| "'" [^ '\'' ]* "'"
	{ 	let val s = Lexing.lexeme lexbuf 
			val v = String.sub s 1 (String.length s - 2) 
		in 	set_line_no(s); (Unparsed_string(v), Within_tag) end }
	| "'"
	{ 	raise XMLParserError(Mismatch("the second quotation mark '", "tag", !line_no)) }
	| '>'
	{ 	if (!tag_count = 0) then (Rangle, Document) 
		else (Rangle, Content) }
	| "/>"
	{ 	tag_count := !tag_count - 1 ;
		if (!tag_count = 0) then (Rangle_empty, Document) 
		else (Rangle_empty, Content) }
	| eof
	{ (Eof, Within_tag) }
	| _
	{  raise XMLParserError(Illegal(Lexing.lexeme lexbuf, "tag", !line_no,"")) } 

(* 4. scan_document_type: after "<!DOCTYPE" until matching ">" *)
and scan_document_type = parse
        "PUBLIC" 
        {
          len_of_1st_entering_ext_subset := (List.length !lexbuf_stack);
          (T_Public,Ext_subset)
        }
        | "SYSTEM" 
        {
          len_of_1st_entering_ext_subset := (List.length !lexbuf_stack);
          (T_System,Ext_subset)
        }        
	| name
	{ 	let val s = Lexing.lexeme lexbuf 
		in (Name s, Document_type) end }
	| ws+
	{  set_line_no(Lexing.lexeme lexbuf); (Ignore, Document_type) }	
	| '['
	{ 	(LBRACK, Declaration) }
	| '>'
	{ 	(Doctype_rangle, Document) }
	| eof
	{  	(Eof, Document_type) }
	| '&' 
	{  raise XMLParserError(Not_allowed("References to general entities", 
		"document type", !line_no)) }
	| '%' 
	{ 	raise XMLParserError(Not_allowed("References to parameter entities", 
		"document type", !line_no))}
	| _
	{ 	raise XMLParserError(Not_allowed((Lexing.lexeme lexbuf)^"(seems to be content)",
		"document type", !line_no))}
	  
	  
(* 4. scan_document_type: after "<!DOCTYPE" name SYSTEM "..." until matching Eof *)
and scan_ext_subset = parse
    ws+
	{  set_line_no(Lexing.lexeme lexbuf); 
	   (Ignore,Ext_subset)
	}
        | "PUBLIC"      {(T_Public,Ext_subset)}
        | "SYSTEM"      {(T_System,Ext_subset)}        
	| '%' name ';'
	{
	  let 
	    val s = (Lexing.lexeme lexbuf)	    
	    val id = (String.sub s 1 (String.length s-2))
	    val p_entity = pe_entity_env_lookup id
	  in
	    if (pe_is_empty(p_entity)) then 
	       (Ignore,Ext_subset) //   grammar ƿ Ÿ ʾƾ .
	    else
	       (PERef(id),Ext_subset)
	  end
	}	  
	| '%'
	{(Percent, Ext_subset)}	
	| '&' 
	{ 	raise XMLParserError(Not_allowed("References to general entities is forbidden", 
		"External subset", !line_no))}
	| name
	{ 	(Name (Lexing.lexeme lexbuf), Ext_subset) }
	| nmtoken
	{ 	(Nametoken (Lexing.lexeme lexbuf), Ext_subset) }
	| '+'
	{ 	(Plus, Ext_subset) }
	| '*'
	{ 	(Star, Ext_subset) }
	| '|'
	{ 	(Bar, Ext_subset) }
	| ','
	{ 	(Comma, Ext_subset) }
	| '?'
	{ 	(Qmark, Ext_subset) }
	| '('
	{ 	(Lparen, Ext_subset) }
	| ')'
	{ 	(Rparen, Ext_subset) }
	| "#" name
	{ 	let val s = Lexing.lexeme lexbuf 
		in ( ( case s of
	   		   "#REQUIRED" => Required
			 | "#IMPLIED"  => Implied
	   		 | "#FIXED"    => Fixed
			 | "#PCDATA"   => Pcdata
			 | _           => raise XMLParserError(Illegal((Lexing.lexeme lexbuf), 
			 	                "declaration", !line_no,""))), Ext_subset )
		end }
	| "<!ELEMENT"
	{ 	(Decl_element, Ext_subset) }
	| "<!ATTLIST"
	{ 	(Decl_attlist, Ext_subset) }
	| "<!ENTITY"
	{ 	(Decl_entity, Ext_subset) }
	| "<!NOTATION"
	{ 	(Decl_notation, Ext_subset) }
	| "<!--" comment_string "-->"
	{ 	let	val s = Lexing.lexeme lexbuf
		in 	set_line_no s; (Comment s, Ext_subset)
		end }
	| "<!--"
	{ 	raise XMLParserError(Illegal("comment","declaration", !line_no,
			"possible reasons are illegal control characters or double hyphens ('--') inside the comment, or missing '-->'")) }
	| "<!["
	{ 	(Conditional_begin, Ext_subset) }
	| "]]>"
	{ 	(Conditional_end, Ext_subset) }
	| "["
	{ 	(LBRACK, Ext_subset) }
	|  "<?" XML 
	{ text_decl_context := Ext_subset;
	  (XMLDecl_begin, XML_decl)
	}
	| "<?" name 
	{ 	let val s = Lexing.lexeme lexbuf
			val v = String.sub s 2 ((String.length s) - 2)
		in pi_decl_context := Ext_subset; (PI_begin v, PI_decl) 
		end }
	| "<?" 
	{  (pi_decl_context := Ext_subset; (PI_NULL_begin, PI_decl)) }
	| '"' [^ '"']* '"'
	{ 	
	  let 
	    val s = Lexing.lexeme lexbuf 
	    val str = (String.sub s 1 (String.length s - 2))
	    
	  in 	
	    set_line_no s;
	    ((Unparsed_string (str)), Ext_subset)
	  end  
	}
	| '"'
	{  raise XMLParserError(Mismatch("the second quotation mark \"", 
		"PI declaration", !line_no)) }
	| "'" [^ '\'']* "'"
	{ 	let val s = Lexing.lexeme lexbuf 
		in 	set_line_no s;
			((Unparsed_string (String.sub s 1 (String.length s - 2))), Ext_subset)
		end }
	| "'"
	{  raise XMLParserError(Mismatch("the second quotation mark '", "PI declaration", 
			!line_no)) }
	| '>'
	{  (Decl_rangle, Ext_subset) }
	| eof
	{  
	   if ((List.length !lexbuf_stack) 
	        = !len_of_1st_entering_ext_subset + 1) then 
	     (Eof, Document_type)
	   else
	     (Eof, Ext_subset)  
	 } 
	| "<!"
	{  raise XMLParserError(Bad_declaration(
		"Declaration either badly formed or not allowed", "declaration", !line_no) ) }
	| _
	{  raise XMLParserError(Illegal((Lexing.lexeme lexbuf), "declaration", !line_no,"")) }
			  
(* 5. scan_declaration: after "[" in DTD until matching "]" *)
(*  | "<!ENTITY"(ws+)(name)(ws+)EntityDef(ws* )">"  | "<!ENTITY"(ws+)"%"(ws+)(name)(ws+)PEDef(ws* )">" *)
and scan_declaration = parse
    ws+
	{ 	set_line_no(Lexing.lexeme lexbuf) ; (Ignore,Declaration)}(*//scan_declaration lexbuf  }*)
        | "PUBLIC" 
        {
          (T_Public,Declaration)
        }
        | "SYSTEM" 
        {
          (T_System,Declaration)
        }        
 	| '%' name ';'
	{
	  let 
	    val s = (Lexing.lexeme lexbuf)	    
	    val id = (String.sub s 1 (String.length s-2))
	    val p_entity = pe_entity_env_lookup id
	  in
	    if (pe_is_empty(p_entity)) then 
	       ( 
	       if (!Global.debug_tag) then 
	          (print_string ("ignoring peref("^id^")");
	          print_newline());
	       (Ignore,Declaration) //   grammar ƿ Ÿ ʾƾ .
	       )
	    else
	       (PERef(id),Declaration)
	  end
	}	
	| '%'
	{ 	(Percent, Declaration) }	
	| '&' 
	{ 	raise XMLParserError(Not_allowed("References to general entities is forbidden", 
		"DTDs", !line_no))}
	| name
	{ 	(Name (Lexing.lexeme lexbuf), Declaration) }
	| nmtoken
	{ 	(Nametoken (Lexing.lexeme lexbuf), Declaration) }
	| '+'
	{ 	(Plus, Declaration) }
	| '*'
	{ 	(Star, Declaration) }
	| '|'
	{ 	(Bar, Declaration) }
	| ','
	{ 	(Comma, Declaration) }
	| '?'
	{ 	(Qmark, Declaration) }
	| '('
	{ 	(Lparen, Declaration) }
	| ')'
	{ 	(Rparen, Declaration) }
	| "#" name
	{ 	let val s = Lexing.lexeme lexbuf 
		in ( ( case s of
	   		   "#REQUIRED" => Required
			 | "#IMPLIED"  => Implied
	   		 | "#FIXED"    => Fixed
			 | "#PCDATA"   => Pcdata
			 | _           => raise XMLParserError(Illegal((Lexing.lexeme lexbuf), 
			 	                "declaration", !line_no,""))), Declaration )
		end }
	| "<!ELEMENT"
	{ 	(Decl_element, Declaration) }
	| "<!ATTLIST"
	{ 	(Decl_attlist, Declaration) }
	| "<!ENTITY"
	{ 	(Decl_entity, Declaration) }
	| "<!NOTATION"
	{ 	(Decl_notation, Declaration) }
	| "<!--" comment_string "-->"
	{ 	let	val s = Lexing.lexeme lexbuf
		in 	set_line_no s; (Comment s, Declaration)
		end }
	| "<!--"
	{ 	raise XMLParserError(Illegal("comment","declaration", !line_no,
			"possible reasons are illegal control characters or double hyphens ('--') inside the comment, or missing '-->'")) }
	| "<!["
	{ 	(Conditional_begin, Declaration) }
	| "]]>"
	{ 	(Conditional_end, Declaration) }
	| "["
	{ 	(LBRACK, Declaration) }
	|  "<?" XML 
	{ text_decl_context := Declaration;
	  (XMLDecl_begin, XML_decl)
	}
	| "<?" name 
	{ 	let val s = Lexing.lexeme lexbuf
			val v = String.sub s 2 ((String.length s) - 2)
		in 
		        pi_decl_context := Declaration;
		        (PI_begin v, PI_decl) 
		end }
	| "<?" 
	{  (pi_decl_context := Declaration; (PI_NULL_begin, PI_decl)) }
	| '"' [^ '"']* '"'
	{ 	
	  let 
	    val s = Lexing.lexeme lexbuf 
	    val str = (String.sub s 1 (String.length s - 2))
	    
	  in 	
	    set_line_no s;
	    (Unparsed_string (str), Declaration)
	  end  
	}
	| '"'
	{  raise XMLParserError(Mismatch("the second quotation mark \"", 
		"PI declaration", !line_no)) }
	| "'" [^ '\'']* "'"
	{ 	let val s = Lexing.lexeme lexbuf 
		in 	set_line_no s;
			((Unparsed_string (String.sub s 1 (String.length s - 2))), Declaration)
		end }
	| "'"
	{  raise XMLParserError(Mismatch("the second quotation mark '", "PI declaration", 
			!line_no)) }
	| '>'
	{  (Decl_rangle, Declaration) }
	| ']'
	{  (Dtd_end, Document_type) }
	| eof
	{  (Eof, Declaration) }
	| "<!"
	{  raise XMLParserError(Bad_declaration(
		"Declaration either badly formed or not allowed", "declaration", !line_no) ) }
	| _
	{  raise XMLParserError(Illegal((Lexing.lexeme lexbuf), "declaration", !line_no,"")) }

(* 6. scan_pi_decl_from_document 						  *) 
(* called when the PI appears in the top level  of ext_document and parsed_entity *)
and scan_pi_decl = parse
	ws+ 		
	{  set_line_no(Lexing.lexeme lexbuf); (Ignore, PI_decl)}
	| pi_string"?>" 
	{ 	let	val s = Lexing.lexeme lexbuf
			val len = String.length s
		in 	set_line_no(s);
			(PI_end( String.sub s 0 (len - 2)), !pi_decl_context)
		end }

(* 7. scan_xml_decl							*)
(* xml declaration in top level						*)
and scan_xml_decl = parse
	ws+	
	{set_line_no(Lexing.lexeme lexbuf); scan_xml_decl lexbuf }
  (* {23} VersionInfo *)
	| "version" Eq "'" VersionNum "'" 				
	{ 	let val s = get_last_squote_str (Lexing.lexeme lexbuf)
	 	in 	set_line_no(Lexing.lexeme lexbuf); (Version s, XML_decl) end }
	| "version" Eq "\"" VersionNum "\""				
	{ 	let val s = get_last_quote_str (Lexing.lexeme lexbuf)
		in	set_line_no(Lexing.lexeme lexbuf); (Version  s, XML_decl) end }

	| "encoding" Eq "'" EncNum "'"					
	{ 	let val s = get_last_squote_str (Lexing.lexeme lexbuf)
	 	in set_line_no(Lexing.lexeme lexbuf); (Encoding s, XML_decl) end }
	| "encoding" Eq "\"" EncNum "\""				
	{ 	let val s = get_last_quote_str (Lexing.lexeme lexbuf)
	 	in  set_line_no(Lexing.lexeme lexbuf); (Encoding s, XML_decl) end }
	| "standalone" Eq "'" ("yes"|"no") "'"		
	{ 	let val s = get_last_squote_str (Lexing.lexeme lexbuf)
	 	in 	set_line_no(Lexing.lexeme lexbuf); (Standalone s, XML_decl) end }
	| "standalone" Eq "\"" ("yes"|"no") "\""		
	{ 	let val s = get_last_quote_str (Lexing.lexeme lexbuf)
	 	in set_line_no(Lexing.lexeme lexbuf); (Standalone s, XML_decl) end }
	| "?>" 
	{ 	(XMLDecl_end,!text_decl_context)}
	| _ 	
	{  raise XMLParserError(Illegal((Lexing.lexeme lexbuf), "XML declaration", 
			!line_no,"")) }

{

(* fun xml_lexer_gen lid_arg                                                    *)
(* - the actual lexer is generated by applying this function to the argument    *)
(*   for the type of files                                                      *)
(* - that is, the lexer is 'xml_lexer_gen lex_id' form                          *)
(* - when lex_id may be one of  Document{01}, Parsed_ent{78} and Pe{79} 	*)

fun xml_lexer_gen lid_arg init_lexbuf  = 
let 
  val debug = !debug_tag
  val _ = push_a_lexbuf(init_lexbuf,true,0)  // this is main lexbuf (start point)
  fun scan_one_token lid lexbuf = 
    (case lid of
     (Document | Parsed_ent) =>  scan_document lexbuf 
    | Document_type => scan_document_type lexbuf
    | Content => scan_content lexbuf
    | Within_tag => scan_within_tag lexbuf
    | (Declaration | Pe) => (scan_declaration lexbuf)
    | XML_decl => scan_xml_decl lexbuf  
    | PI_decl => scan_pi_decl lexbuf
    | Ext_subset => scan_ext_subset lexbuf)
    
  fun lid_string lid =
    (case lid of
      Document => "document"
    | Parsed_ent =>  "parsed_entity"
    | Document_type => "document_type"
    | Content => "content"
    | Within_tag => "within tag"
    | Declaration => "declaration"
    | Pe => "pe"
    | XML_decl => "xml_decl" 
    | PI_decl => "pi_decl"
    | Ext_subset => "ext_subset"
    )
  fun print_lid(lid) = print_string ("(lid:"^(lid_string lid)^")")
   
  val t1 = lid_arg
  val lex_id = ref t1
    
  fun scan_next lexbuf = let 
    val (tok,lid) = scan_one_token (!lex_id) (current_lexbuf()) 
  in
    if (tok = Ignore) then 
      (lex_id := lid;           
       scan_next lexbuf)
    else if (tok = Eof) then
      if (List.length !lexbuf_stack > 1) then
          (lex_id := lid; 
           if (!Global.debug_tag) then
             (print_string "EOF with lid = "; 
              print_string (lid_string lid); 
              print_newline());
           if (is_current_lexbuf_external()) then
               (if (!Global.debug_tag) then 
                   print_string "popping a uri\n";
                pop_off_a_URI());
           line_no := pop_off_a_lexbuf();     
           scan_next lexbuf)
      else    
          Eof
    else 
      (lex_id := lid;
       tok)
  end

  fun new_scan_next lexbuf = 
    let       
      val tok = scan_next lexbuf
    in
    	if debug = true then (
    	  print_tok tok;
    	  print_lid (!lex_id)
    	);  
    	tok
    end
in
  new_scan_next
end
}


