(*********************************************************************)
(*                nxml.n                                             *)	
(*-------------------------------------------------------------------*)
(*  by Hyun-Goo Kang                                                 *)
(*-------------------------------------------------------------------*)
(* Define all api to process a XML document                          *)
(*  - this is the interface with a nML application                   *)
(*    which use NXML processor                                       *)
(*********************************************************************)

signature NXML = 
sig
  type document_types = 
              DT_DOCUMENT             // a normal XML document to start
            | DT_EXT_PARSED_ENT       // an external subset of data type definition
            | DT_EXT_PE               // an external subset of content  
  type printing_tag = 
              PRINT_ROOT_ONLY         // pretty printing root element only
            | PRINT_ALL               // pretty printing all

  // parse given XML document 
  val parse : string                  // a uri
           -> document_types 
           -> Nxml_ast.t_entry_point
  
  // pretty printing of abstract syntax similarly with the source document
  val print_absyn : Nxml_ast.t_entry_point
           -> printing_tag 
           -> unit
  
  // print error message if there is validation error
  val validate : Nxml_ast.t_entry_point -> unit 
           
  // for debugging only (for hgkang)         
  val set_debug_mode : unit -> unit    
end


structure NXML : NXML = 
struct

type document_types = DT_DOCUMENT | DT_EXT_PARSED_ENT | DT_EXT_PE
type printing_tag = PRINT_ROOT_ONLY | PRINT_ALL

fun parse uri doct = let    
  val absolute= System_Processing.absolute_path (Global.current_URI()) uri
  val _ = Global.push_a_URI absolute 
  val ducument_string = System_Processing.string_from_system_literal absolute 
  val lexbuf = Lexing.from_string ducument_string
  fun parse lexbuf =
    case doct of 
      DT_DOCUMENT => Parser.document (Lexer.xml_lexer_gen Lexer.Document lexbuf) lexbuf
    | DT_EXT_PARSED_ENT => Parser.extParsedEnt (Lexer.xml_lexer_gen Lexer.Parsed_ent lexbuf) lexbuf
    | DT_EXT_PE => Parser.extPE (Lexer.xml_lexer_gen Lexer.Pe lexbuf) lexbuf
    
in 
  parse lexbuf 
  handle Nxml_errors.XMLParserError(t) => (Parsing.clear_parser () ; 
                                           Nxml_errors.print_error_msg(t); 
                                           exit(1) )
       | x  => (Parsing.clear_parser(); raise x )  
end       

fun print_absyn absyn tag = 
  case tag of 
    PRINT_ROOT_ONLY => Nxml_ast.print_root_elm_only absyn
  | PRINT_ALL => Nxml_ast.print_ast absyn 
  
fun validate absyn = Validate.validate absyn 
    handle Validate.VC(msg) => (print_string ("\nInvalid document: "^msg); 
                                print_newline()) 

fun set_debug_mode () = (Global.debug_tag := true)

end  // structure
