(*********************************************************************)
(*                system.n                                           *)	
(*-------------------------------------------------------------------*)
(*  by Hyun-Goo Kang                                                 *)
(*-------------------------------------------------------------------*)
(*  Implement system access related processing                       *)
(*   - file processing                                               *)
(*   - http processing                                               *)
(*   - absolute path calculation                                     *)
(*   - string normalization                                          *)
(*      . see subsection 2.11 in XML spec: End-of-Line Handling      *) 
(*      . preprocessed                                               *)
(*********************************************************************)

signature SYSTEM_PROCESSING =
sig 
  val string_from_system_literal : 
                    string    // absoulte uri
                 -> string    // target string (normalized)  
  val absolute_path : string // current ducument's uri (absolute)
                 -> string   // new uri string (can be relative)
                 -> string   // new uri's absolute uri
end

structure System_Processing : SYSTEM_PROCESSING = 
struct 

open Global
open Unix
open Str
//exception Http of string
exception Exn_system of string

fun normalize_string str = let  
  val pat_str = "\013\010\|013"
  val pat = Str.regexp pat_str
in  
  Str.global_replace pat "\010" str
end handle Not_found => str

// url => ip 
fun ip_of_url url = let
  val host_entry = gethostbyname url handle Not_found => raise Exn_system("http: can't find the server: " ^ url)
  val h_addr = host_entry.h_addr_list
in h_addr.[0] end

// recv stream 
fun receive_stream(sockfd) = let
  val buffer_len = 10
  fun read_sock buffer = let
    val temp_buffer = String.make buffer_len ' '
    val numbytes = recv sockfd temp_buffer 0 buffer_len [] 
                   handle Unix.Unix_error( enum, ecmd, arg) => 
                          raise Exn_system("HTTP: Fail to receive.\n")
    in 
      if (numbytes > 0) then read_sock (buffer ^ temp_buffer)
      else buffer
    end
in read_sock (String.make 0 ' ') end 

fun get_header post host content_len = let
      val http_ver = "1.0"
      val content_type = "text/xml" (* must be text/xml *)
      val charset="utf-8"
  in 
      "POST " ^ post ^ " HTTP/" ^ http_ver ^ "\r\nHost: " ^ host^ "\r\nContent-Type: " ^ content_type ^ "; charset=" ^ charset ^ "\r\nContent-Length: " ^ (string_of_int content_len) ^ "\r\nSOAPAction: \"\"\r\n\r\n" 
end

fun http_post host_url port header body = let
  val host_ip = ip_of_url host_url
  val sockfd = ref (socket PF_INET SOCK_STREAM 0)
  val target_addr = ref ( ADDR_INET( host_ip, port ) ) 
  // try to make connection  
  val _ = connect !sockfd !target_addr 
          handle Unix.Unix_error( enum, ecmd, arg ) => 
                 raise Exn_system("HTTP: Receiver is not ready to get your call.\n")  
  // send header stream 
  val header_sent = send !sockfd header 0 (String.length header) []
                    handle Unix.Unix_error( enum, ecmd, arg) => 
                           raise Exn_system("HTTP: Fail to send header\n") 
  // send body stream 
  val body_sent = send !sockfd body 0 (String.length body) []
               handle Unix.Unix_error( enum, ecmd, arg) => 
                      raise Exn_system("HTTP: Fail to send body\n")
  // recv stream 
  val buff = receive_stream(!sockfd)  

  // flush socket 
  val _ = shutdown !sockfd SHUTDOWN_ALL  	
  
  fun split_msg (buff) = let
    val split_list = split (regexp "\r\n\r\n") buff
  in
    (List.hd split_list, List.hd (List.tl split_list))
  end 
in
  split_msg(buff) // (header, body)
end

fun http_get hostname port target' = let
  val target = "GET" ^ " "  ^ target' ^ "\r\n\r\n"
  // init 
  val host_ip = ip_of_url hostname
  val sockfd = (socket PF_INET SOCK_STREAM 0)
  val target_addr = ( ADDR_INET( host_ip, port ) ) 

  // try to make connection  
  val _ = connect sockfd target_addr 
           handle _ => //Unix.Unix_error( enum, ecmd, arg ) => 
                  raise Exn_system("HTTP: Can't connect \"" ^ hostname 
                              ^ "using port " ^ (string_of_int port))
  // send get stream
  val get_msg_sent = send sockfd target 0 (String.length target) []
           handle Unix.Unix_error( enum, ecmd, arg) => 
                  raise Exn_system("HTTP: Fail to send HTTP msg: " ^ target)

  // recv stream 
  val buff = receive_stream(sockfd)
  val _ = shutdown sockfd SHUTDOWN_ALL  // flush socket 
in
  buff
end
         
fun get_string_from_file(filename) = let
  val ic = open_in filename
  fun read_ic buffer = let
    val temp_buffer = String.make 10 ' '
    val numbytes = input ic temp_buffer 0 10
    in 
      if (numbytes > 0) then read_ic (buffer ^ temp_buffer)
      else buffer
    end 
  val str = read_ic ""   
  val _ = close_in ic 
in str  end 


fun absolute_path c_uri_str sys_str = let
  val scheme = Neturl.extract_url_scheme sys_str 
               handle Neturl.Malformed_URL => "relative_uri"
in 
  case scheme of 
    ("http"|"file") => sys_str
  | "relative_uri" => 
    let val cur_scheme = Neturl.extract_url_scheme c_uri_str
    in         
     (case cur_scheme of 
       "file" =>
         let
           val file = Hashtbl.find Neturl.common_url_syntax "file"           
           val r_u = Neturl.url_of_string file ("file:"^sys_str)
           val c_u = Neturl.url_of_string file (c_uri_str)
           val r_path = Neturl.url_path r_u
           val c_path = Neturl.url_path c_u
           val new_path = List.rev(List.tl(List.rev(c_path)))@r_path                
           val filename = Neturl.join_path (Neturl.norm_path new_path)
           val new_uri_str = filename
         in
           ("file:"^new_uri_str)
         end
       | "http" => 
         let 
           val http = Hashtbl.find Neturl.common_url_syntax "http"
           val c_u = Neturl.url_of_string http c_uri_str 
           val c_p = Neturl.url_path c_u handle Not_found => [""]
           val c_p' = List.rev(List.tl(List.rev c_p))           
           val r_p = Neturl.split_path sys_str 
           val r_p' = Neturl.norm_path r_p
           val a_p = Neturl.join_path(c_p'@r_p')
           val absolute_str = "http://" ^
                        (Neturl.url_host c_u) ^ ":" ^
                        (string_of_int((Neturl.url_port c_u) handle Not_found => 80)) ^ 
                        a_p          
        in
          absolute_str
        end                
       | _ => raise Exn_system("non-recognizable relative URI")
    )   
    end
  | _ => ("external beyond http or file:" ^ sys_str)  
end handle x => (print_string ("fail to get absoulte uri from " ^ sys_str ^"\n");
                 raise x)


fun string_from_system_literal sys_str = let   // currently http, file check only
  val _ = if (!Global.debug_tag) then 
            (print_string ("\nstring from url :\ncalled with " ^ "\""); 
             print_string (sys_str ^ "\" "); print_newline() )
  val scheme = Neturl.extract_url_scheme sys_str 
               handle Neturl.Malformed_URL => "relative_uri"
                   | _ => (print_string "neturl library exception"; "")
in 
  case scheme of 
    "http" => 
    let 
      val http = Hashtbl.find Neturl.common_url_syntax "http"
      val u = Neturl.url_of_string http sys_str        	  
      val host = Neturl.url_host u         	  
      val port = Neturl.url_port u handle Not_found => 80 
      val url_string = Neturl.string_of_url u
    in
      normalize_string(http_get host port url_string)
    end
  | "file" => 
    let
      val file = Hashtbl.find Neturl.common_url_syntax "file"
      val u = Neturl.url_of_string file sys_str
      val host = Neturl.url_host u handle Not_found => ""
      val path = Neturl.url_path u handle Not_found => [""]
      val filename = Neturl.join_path path
    in
      normalize_string(get_string_from_file(filename))
    end                 
  | _ => ("external beyond http or file:" ^ sys_str)
end handle x => (print_string("fail to get external entity from " ^ sys_str);
                 print_newline(); raise x)

end
